#ifndef MICROPY_INCLUDED_TEENSY_LEXERMEMZIP_H
#define MICROPY_INCLUDED_TEENSY_LEXERMEMZIP_H

mp_lexer_t *mp_lexer_new_from_memzip_file(const char *filename);

#endif // MICROPY_INCLUDED_TEENSY_LEXERMEMZIP_H
